/*
 * MASC (Mobitex Asynchronous Communication Protocol)
 * Protocol Implementation for Linux
 * Copyright 1998 Kaz Kylheku <kaz@cafe.net>
 * This work is offered under the GNU Public License
 * as part of the Linux operating system.
 */

/*
 * $Id: mobitex_masc.c,v 1.53 2000/02/11 05:02:25 kaz Exp $
 * $Name: masc_2_6 $
 */

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/config.h>
#include <linux/module.h>
#include <linux/if.h>
#include <linux/if_ether.h>
#include <linux/if_arp.h>
#include <linux/skbuff.h>
#include <linux/inet.h>
#include <linux/timer.h>
#include <linux/proc_fs.h>
#include <asm/byteorder.h>
#include <asm/uaccess.h>
#include <net/arp.h>
#include <linux/ctype.h>
#include "mobitex_masc.h"
#include "mobitex_masc_ioctl.h"

#ifdef MODULE
#define MASC_VERSION    "2.6 20000210 (MODULE)"
#else
#define	MASC_VERSION	"2.6 20000210"
#endif

#define MNAME "MOBITEX_MASC: "

/*
 * Some pre-cooked useful MASC frames, complete with checksums.
 * Why don't I use string literals here? Because these have
 * a null byte at the end which adds 1 to the size, which forces
 * you to either write the size in the array declarator, or
 * subtract 1 all over the place, two error-prone alternatives. :)
 */

static char masc_init_frame[] = {
	'^', '0', '0', '1', '0', 'B', ' ', '4', '7', 'E', ',', '0', ':', '5', 'D', '\r'
};

static char masc_getman_frame[] = {
	'^', '0', '0', '0', 'C', 'F', ' ', 'P', ':', '2', '1', '\r'
};

static char masc_getbatt_frame[] = {
	'^', '0', '0', '0', 'E', 'F', ' ', 'A', '0', '3', ':', '3', '5', '\r'
};

static char masc_getsign_frame[] = {
	'^', '0', '0', '0', 'E', 'F', ' ', 'A', '0', '2', ':', '3', '4', '\r'
};

static char masc_turnoff_frame[] = {
	'^', '0', '0', '0', 'C', 'F', ' ', 'O', ':', '3', 'E', '\r'
};

static char masc_xmit_off_frame[] = {
	'^', '0', '0', '0', 'D', 'F', ' ', 'M', '0', ':', '0', 'B', '\r'
};

static char masc_xmit_on_frame[] = {
	'^', '0', '0', '0', 'D', 'F', ' ', 'M', '1', ':', '0', 'A', '\r'
};

static int masc_maxdev = MASC_NRUNIT;	/* max number of MASC devices	*/
static volatile int masc_numdev = 0;	/* counter for naming devices	*/
static int masc_proc_flag;		/* proc filesystem registration	*/

static masc_dev_t masc_list;	/* circular doubly linked list of MASCs */

static struct tty_ldisc	masc_ldisc;	/* line discipline operations	*/
static struct proc_dir_entry masc_proc;	/* proc filesystem entry	*/

typedef enum { MASC_SMALL, MASC_BIG } masc_size_t;

static const int masc_size[2] = { MASC_SMALL_BUF_SIZE, MASC_BIG_BUF_SIZE };

/*
 * MAN numbers to send broadcasts to. We don't want to do multicasts over the
 * Mobitex network. But for now we can direct a broadcast at the station of our
 * choice using the mechanism below.  By selecting a designated host which is
 * set up with proxy ARP entries, we can use ARP to resolve IP addresses to MAN
 * numbers.  Note that this is probably a bad idea; it makes more sense for
 * mobile units to be set up with fixed arp caches. These numbers can be set
 * when loading the driver, for example: insmod mobitex_masc
 * masc_bcast=5000123,6000124 to set up masc0 to
 * send broadcasts to MAN 5000123 and masc1 to send broadcasts to 6000124.
 */

int masc_bcast[MASC_NRUNIT];
MODULE_PARM(masc_bcast, "1-" __MODULE_STRING(MASC_NRUNIT) "i");


/*
 * global driver spinlock
 */

#ifdef __SMP__	/* prevent warning about unused */
static spinlock_t masc_spinlock;
#endif

static void masc_global_lock(void);
static void masc_global_unlock(void);
static void masc_lock(masc_dev_t *);
static void masc_unlock(masc_dev_t *);
static struct sk_buff *masc_alloc_buf(masc_dev_t *, masc_size_t, int);
static void masc_free_buf(masc_dev_t *, struct sk_buff *);
static void masc_load_tx_buf(masc_dev_t *, struct sk_buff *);
static void masc_cleanout_old(masc_dev_t *);
static int masc_fill_transmit_buf(masc_dev_t *);
static void masc_kick_tty(masc_dev_t *);
static long masc_read_hex(unsigned char *, int);
static int masc_infoframe_basic_validate(unsigned char *, int);
static int masc_infoframe_validate(unsigned char *, int);
static void masc_send_ctrl(masc_dev_t *, char *);
static void masc_send_ack(masc_dev_t *, int);
static void masc_process_ack(masc_dev_t *, unsigned char *, int);
static void masc_process_nack(masc_dev_t *, unsigned char *pak, int);
static struct sk_buff *masc_get_untransmitted(masc_dev_t *, int);
static void masc_send_info(masc_dev_t *, char *, int, int);
static void masc_cksum_send_info(masc_dev_t *, char *, int, int);
static int masc_modem_off(masc_dev_t *);
static int masc_begin_ioctl(masc_dev_t *, struct wait_queue *);
static int masc_ioctl_wait(masc_dev_t *, struct wait_queue *);
static void masc_end_ioctl(masc_dev_t *, struct wait_queue *);
static int masc_transmitter_off(masc_dev_t *);
static int masc_transmitter_on(masc_dev_t *);
static int masc_set_netid(masc_dev_t *, unsigned long);
static int masc_wait_drain(masc_dev_t *);
static void masc_mpak_to_ether(struct sk_buff *);
static void masc_ether_to_mpak(struct sk_buff *);
static struct sk_buff *masc_mframe_decap(masc_dev_t *, unsigned char *, int);
static void masc_process_mframe(masc_dev_t *, unsigned char *, int);
static void masc_handle_error(masc_dev_t *, int, int);
static void masc_send_init(masc_dev_t *);
static int masc_process_infoframe(masc_dev_t *, unsigned char *, int);
static void masc_process_frame(masc_dev_t *);
static void masc_free_buffers(masc_dev_t *);
static int masc_dev_open(struct device *);
static int masc_dev_ioctl(struct device *, struct ifreq *, int);
static int masc_dev_stop(struct device *);
static void masc_encaps_mframe(masc_dev_t *, struct sk_buff *, struct sk_buff *);
static int masc_dev_xmit(struct sk_buff *, struct device *);
static struct net_device_stats *masc_dev_stats(struct device *);
static int masc_dev_hard_header(struct sk_buff *, struct device *, unsigned short, void *, void *, unsigned int);
static int masc_dev_rebuild_header(struct sk_buff *);
static int masc_mac_addr(struct device *, void *);
static void masc_ack_timer(unsigned long);
static void masc_bat_timer(unsigned long);
static void masc_traf_timer(unsigned long);
static void masc_off_timer(unsigned long);
static void masc_init_timer(unsigned long);
static int masc_dev_init(struct device *);
static masc_dev_t *masc_find_free(void);
static masc_dev_t *masc_dev_alloc(void);
static void masc_dev_dealloc(masc_dev_t *);
static int masc_tty_open(struct tty_struct *);
static void masc_tty_close(struct tty_struct *);
static int masc_tty_ioctl(struct tty_struct *, struct file *, unsigned int, unsigned long);
static void masc_tty_receive_buf(struct tty_struct *, const unsigned char *, char *, int);
static int masc_tty_receive_room(struct tty_struct *);
static int masc_get_info(char *, char **, off_t, int, int);
static void masc_tty_write_wakeup(struct tty_struct *);
static int masc_do_init(void);

/*
 * locking functions to ensure that ensure only one processor
 * gets access.
 */

static void masc_global_lock(void)
{
	spin_lock_irq(&masc_spinlock);
}

static void masc_global_unlock(void)
{
	spin_unlock_irq(&masc_spinlock);
}

static void masc_lock(masc_dev_t *masc)
{
	long flags;
	spin_lock_irqsave(&masc->spinlock, flags);
	masc->flags = flags;
}

static void masc_unlock(masc_dev_t *masc)
{
	long flags = masc->flags;
	spin_unlock_irqrestore(&masc->spinlock, flags);
}

/*
 * Allocate a new buffer for a MASC frame. Try to get this from one of the two
 * free lists maintained by the device, else allocate it dynamically. A null
 * pointer is returned on failure. Called with device structure locked.
 */

static struct sk_buff *masc_alloc_buf(masc_dev_t *masc, masc_size_t size, int priority)
{
	struct sk_buff *skb = __skb_dequeue(&masc->free_queue[size]);

	if (skb == NULL) {
		masc_unlock(masc);
		skb = alloc_skb(masc_size[size], priority);
		masc_lock(masc);
	} else {
		__skb_trim(skb, 0);
	}

	return skb;
}

/*
 * Free buffer by putting it on one of the two free lists depending
 * on its size. Called with masc_dev_t locked.
 */

static void masc_free_buf(masc_dev_t *masc, struct sk_buff *skb)
{
	masc_size_t size = MASC_SMALL;
	if (skb->truesize >= MASC_BIG_BUF_SIZE)
		size = MASC_BIG;
	__skb_queue_tail(&masc->free_queue[size], skb);
}

/*
 * Load contents of socket buffer into transmit buffer.  Called from
 * masc_fill_transmit_buf which ensures serialization so we don't
 * need to worry about concurrent accesses to transmit buffer.
 * Called with device structure locked.
 */

static void masc_load_tx_buf(masc_dev_t *masc, struct sk_buff *skb)
{
	memcpy(masc->tx_buf, skb->data, skb->len);
	masc->tx_left = skb->len;
	masc->tx_head = masc->tx_buf;
}

/*
 * Scan the FH queue and clean out any old packets---ones that have not
 * received their FH acknowledgement for too long. This keeps the
 * queue from being plugged up.
 * Called with device structure locked.
 */

static void masc_cleanout_old(masc_dev_t *masc)
{
	int qlen;

	qlen = skb_queue_len(&masc->fh_queue);

	while (qlen--) {
		struct sk_buff *skb = __skb_dequeue(&masc->fh_queue);
		masc_skb_cb_t *masc_cb = (masc_skb_cb_t *) skb->cb;
		if (jiffies - masc_cb->timestamp > MASC_TRANS_TIMEOUT*HZ) {
			masc_free_buf(masc, skb);
			masc->net_stats.tx_dropped++;
		} else {
			__skb_queue_head(&masc->fh_queue, skb);
		}
	}
}

/*
 * Do whatever it takes to stuff the transmit buffer. Return 0 to indicate that
 * there is nothing in the transmit buffer, otherwise return 1.
 * Called with device locked.
 */

static int masc_fill_transmit_buf(masc_dev_t *masc)
{
	struct sk_buff *skb;

	/*
	 * The transmit buffer has data, we have nothing to do.
	 */

	if (masc->tx_left != 0)
		return 1;

	/*
	 * The control queue has the highest priority, so we check it first.
	 * Since control packets don't have to be retained for retransmission,
	 * once a control packet is sent, we can simply throw its buffer onto
	 * the queue of free buffers. The retransmit logic will create these
	 * packets from scratch.
	 */

	if ((skb = __skb_dequeue(&masc->ctrl_queue)) != NULL) {
		masc_load_tx_buf(masc, skb);	
		masc_free_buf(masc, skb);
		return 1;
	} 

	/*
	 * Next, we check whether we are waiting for an ACK/NACK from the radio
	 * modem by checking whether the ack wait queue has a packet in it.  If
	 * we are waiting for an ACK, we cannot slide down any of the
	 * information frames, so we must sadly leave the transmit buffer
	 * empty---unless there is a request to retransmit the packet that is
	 * waiting for the ack. If we retransmit the packet, we must also
	 * enable the generation of RACKs, restart the RACK timer.
	 */

	if (skb_queue_len(&masc->ackw_queue) > 0) {
		if (masc->flg_retran) {
			masc->flg_retran = 0;
			skb = skb_peek(&masc->ackw_queue);
			if (skb->protocol == MASC_INIT)
				masc->flg_initack = 1;
			masc_load_tx_buf(masc, skb);
			del_timer(&masc->ack_timer);
			masc->ack_timer.expires = jiffies + MASC_ACK_TIMEOUT*HZ;
			add_timer(&masc->ack_timer);
			return 1;
		}
		return 0;
	}

	/*
	 * There is nothing requiring an ack, so kill the timer in
	 * case it is running---which it should not be, since either
	 * this is the very first time through this code, or we
	 * got this far because an ack was successfully received,
	 * in which case the timer would have been killed.
	 */

	del_timer(&masc->ack_timer);

	/*
	 * Next, we check whether we have received an INIT frame from the modem.
	 * If not, we don't send anything to it. Other logic in the driver
	 * ensures that the INIT frame is the first one we send to the modem.
	 */

	if (!masc->flg_initsent)
		return 0;

	/*
	 * Clear the bit saying that we have sent a rack, since
	 * we will be starting on a brand new information frame.
	 */

	masc->flg_racks = 0;
	masc->ack_timer.expires = jiffies + MASC_ACK_TIMEOUT*HZ;

	/*
	 * We now check the queue of non-MPAK info frames first, giving it
	 * higher priority.  This queue could contain an INIT frame, or
	 * commands to sense the battery status, etc. None of the packets
	 * in this queue are destined for radio transmission; they are strictly
	 * between Linux and the radio modem.
	 */
	 
	if ((skb = __skb_dequeue(&masc->info_queue)) != NULL) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: queuing info frame\n", masc->if_name));
		if (skb->protocol == MASC_INIT)
			masc->flg_initack = 1;
		__skb_queue_tail(&masc->ackw_queue, skb);
		masc_load_tx_buf(masc, skb);
		add_timer(&masc->ack_timer);
		return 1;
	}

	/*
	 * Before we check the MPAK queue, we also check whether we have have
	 * been asked to shut up at the request of the Mobitex network. If so,
	 * we politely refrain from sending any packets to the radio modem.
	 */

	if (masc->flg_tstop) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: throttled by command from network\n", masc->if_name));
		return 0;
	}

	/* 
	 * Rid the FH queue of stale packets.
	 */

	masc_cleanout_old(masc);

	/*
	 * We check whether the queue of M frames which have not yet received
	 * an FH acknowledgement frame has reached the maximum. If not, we can
	 * proceed, and try to grab the next M frame from the MPAK queue and
	 * stuff it into the modem. When that packet is acked, the
	 * masc_process_ack() function will add it to the fh_queue, where it
	 * will await an FH frame from the radio modem, indicating that it has
	 * been transmitted and removed from the radio modem's own internal
	 * queue. MASC_MAX_FRAME is the maximum number of unradioed M frames
	 * that the modem can supposedly hold. If the fh_queue is currently
	 * full, we must return empty handed (or empty buffered, so to speak).
	 */	

	if (skb_queue_len(&masc->fh_queue) >= MASC_MAX_MFRAME) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: max outstanding M frames reached\n", masc->if_name));
		return 0;
	}

	if (masc->traf_count >= MASC_MAX_AIRED_PACKETS) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: max aired packets reached, backing off\n", masc->if_name));
		return 0;
	}

	/*
	 * Try to get a frame from the MPAK queue. If it is empty, clear
	 * the busy flag to indicate that the device can take more data
	 * and return.  But don't clear the busy flag if we have not yet
	 * received our local MAN---we want to make the device appear busy
	 * until we have learned of our hardware address, so that we
	 * never transmit packets with a bogus source MAN.
	 */

	if ((skb = __skb_dequeue(&masc->mpak_queue)) == NULL) {
		if (masc->flg_haveman) {
			masc->dev->tbusy = 0;
			mark_bh(NET_BH);
		}
		if (waitqueue_active(&masc->drain_wait))
			wake_up_interruptible(&masc->drain_wait);
		return 0;
	}

	/* 
	 * We have an MPAK frame to send, so we load it into the transmit
	 * buffer and put it into the one-element ACK waiting queue,
	 * then start the RACK timer.
	 */

	MASC_DBG_PRINTK((KERN_INFO MNAME "%s: sending M frame\n", masc->if_name));
	__skb_queue_tail(&masc->ackw_queue, skb);
	masc_load_tx_buf(masc, skb);
	add_timer(&masc->ack_timer);
	return 1;
}

/*
 * Send data down to tty to initiate data flow, if it is not transmitting
 * already! This should be called whenever data has been made available on the
 * queues. Called with device locked.
 */

static void masc_kick_tty(masc_dev_t *masc)
{
	struct device *dev = masc->dev;

	if (masc->flg_pushing) { 	/* prevent being reentered by tty receive interrupt */
		MASC_DBG_PRINTK((KERN_INFO MNAME "someone already pushing data out of %s\n", masc->if_name));
		return;
	}

	masc->flg_pushing = 1;

	/*
	 * We check whether we have not been able to transmit the packet
	 * for 20 seconds or longer, in which case we report an error and toss it.
	 */

	if (masc->tx_left > 0 && jiffies - dev->trans_start >= 20 * HZ) {
		dev->trans_start = jiffies;
		printk(KERN_ERR MNAME "transmit timed out for %s\n", masc->if_name);
		masc->net_stats.tx_errors++;
		masc->tx_left = 0;
	}

	/*
	 * We have now marked the device as busy, and so we test whether we can
	 * deliver data.  We might not be able to, even though there are queued
	 * frames, because we are waiting for an ack, or are out of contact
	 * with the network, etc.  If we can't deliver data, we must clear the
	 * busy flag and disable write wakeup callbacks.  Otherwise, we enable
	 * write wakeups and stuff the data into the tty device to kick-start
	 * the downward flow.
	 */

	/*
	 * Loop until there is no more data in the queues that can be
	 * transmitted immediately, or until the tty cannot take any more.
	 */

	for (;;) {
		if (!masc_fill_transmit_buf(masc)) {
			MASC_DBG_PRINTK((KERN_INFO MNAME "%s: end of write burst\n", masc->if_name));
			masc->tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
			break;
		} else {
			int nwrit;
			struct tty_struct *tty = masc->tty;
			tty->flags |= (1 << TTY_DO_WRITE_WAKEUP);
			masc_unlock(masc);
			/* pushing flag ensures nobody touches tx_ stuff while we are unlocked */
			nwrit = tty->driver.write(tty, 0, masc->tx_head, masc->tx_left);
			masc_lock(masc);
			masc->tx_left -= nwrit;
			masc->tx_head += nwrit;
			/* if we were able to write something, set the last transmit time */
			if (nwrit)
				masc->dev->trans_start = jiffies;
			if (masc->tx_left != 0)
				break;
		}
	}

	masc->flg_pushing = 0;
}

/*
 * Read the specified number of hexadecimal characters from the given buffer,
 * and convert to a long value.
 */

static long masc_read_hex(unsigned char *buf, int len)
{
	long hex = 0;

	while (len--) {
		int ch = *buf++;
		if (isdigit(ch)) {
			ch -= '0';
		} else {
			char *h = "ABCDEF";
			char *p = strchr(h, ch);
			if (!p)
				return -1;
			ch = p - h + 10;
		}
		hex <<= 4;
		hex |= ch;
	}

	return hex;
}


/*
 * Check an incoming MASC information or init packet for basic criteria:
 * - starts with ^ character, indicating it's an info frame
 *   (assumed to be checked by the caller).
 * - contains exactly one colon (:) character;
 * - has a total length of at least 10 characters;
 * - ends with a carriage return (assumed checked by the caller).
 * An information packet which does not meet these criteria is
 * discarded without. An information packet which meets these
 * criteria but does not meet the full criteria requires
 * a NACK response, according to the MASC protocol.
 */

static int masc_infoframe_basic_validate(unsigned char *buf, int len)
{
	unsigned char *p;

	if (len < 10)
		return 0;

	if (!(p = memchr(buf, ':', len)))
		return 0;

	if (memchr(p + 1, ':', len - (p + 1 - buf)))
		return 0;

	return 1;
}

/*
 * Check an incoming MASC information or init packet for the full required
 * criteria:
 * - starts with ^ character;
 * - contains exactly one colon (:) character;
 * - ``check'' and ``length'' fields have correct values;
 * - no character with parity error exists;
 * - maximum number of characters has not been exceeded in any
 *   field or frame;
 * - ends with carriage return.
 * It is assumed that the packet has already passed the basic
 * criteria checked by masc_infoframe_basic_validate, and that the
 * parity errors are checked elsewhere. So we just verify the
 * checksums and sizes.
 */

static int masc_infoframe_validate(unsigned char *buf, int len)
{
	long length;
	int cksum = 0;

	length = masc_read_hex(buf + 1, 4);

	if (length != len) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "validation failed due to bad length\n"));
		return 0;
	}

	while (len-- > 3)
		cksum ^= *buf++;

	MASC_DBG_PRINTK((KERN_INFO MNAME "computed cksum is %x\n", (unsigned) cksum));
	MASC_DBG_PRINTK((KERN_INFO MNAME "scanned cksum is %x\n", (unsigned) masc_read_hex(buf, 2)));
	return cksum == masc_read_hex(buf, 2);
}

/*
 * Send a generic control frame. The given string provides the one
 * or two character packet contents. Since the contents are variable,
 * we must compute the checksum field. It is required that
 * strlen(pkt) <= 2. The device must be locked.
 */

static void masc_send_ctrl(masc_dev_t *masc, char *pkt)
{
	struct sk_buff *skb;
	unsigned char *out;

	skb = masc_alloc_buf(masc, MASC_SMALL, GFP_ATOMIC);

	if (!skb) {
		printk(KERN_ERR MNAME "%s: unable to send control frame due to insufficient memory!\n", masc->if_name);
		return;
	}

	out = skb->data;

	*out++ = '^';
	(void) ((*out++ = *pkt++) && (*out++ = *pkt++));
	*out++ = '\r';

	__skb_put(skb, out - skb->data);
	skb->protocol = MASC_CTRL;
	__skb_queue_tail(&masc->ctrl_queue, skb);
}

/*
 * Send the given ack. The ack parameter must be -1, 0, or 1,
 * and gets converted to the characters '-', '0' or '1'. An
 * ack of '-' means that we have been asked to re-ack (via a RACK
 * frame) but we have never sent an ACK before.
 * Device is locked.
 */

static void masc_send_ack(masc_dev_t *masc, int ack)
{
	char ackstr[3];

	ackstr[0] = '*';
	ackstr[1] = ack["-01"+1];
	ackstr[2] = 0;

	masc_send_ctrl(masc, ackstr);
}

/*
 * Process an ACK frame received from the modem, according to
 * the rules of the MASC protocol. Device must be locked on entry.
 */

static void masc_process_ack(masc_dev_t *masc, unsigned char *pak, int len)
{
	int repeat_frame = 0;

	if (len != 4)
		return;

	masc->nack_count = 0;

	if (masc->flg_racks) {
		/*
		 * INIT frame must be repeated unconditionally if the modem
		 * failed to ACK it and we generated a RACK. This is a special
		 * case required by the MASC protocol specification. Otherwise
		 * the frame is repeated if we get the wrong ACK sequence
		 * number: either - or 1 when expecting 0, or - and 0 when
		 * expecting 1.
		 */
		if (masc->flg_initack) {
			repeat_frame = 1;
		} else {
			switch (pak[2]) {
			case '0':
				if (masc->last_ack != 1)
					repeat_frame = 1;
				masc->last_ack = 0;
				break;
			case '1':
				if (masc->last_ack != 0)
					repeat_frame = 1;
				masc->last_ack = 1;
				break;
			case '-':
				repeat_frame = 1;
				break;	
			default:
				MASC_DBG_PRINTK((KERN_INFO MNAME "ack frame garbled\n"));
				break;
			}
		}
	} else {
		if (pak[2] == '0' || pak[2] == '1')
			masc->last_ack = pak[2] - '0';
	}

	masc->flg_racks = 0;

	if (repeat_frame) {
		masc->flg_retran = 1;
		MASC_DBG_PRINTK((KERN_INFO MNAME "ack resulting in retransmit\n"));
	} else {
		struct sk_buff *skb;

		/*
		 * atomically move ack packet to either free_queue or fh_queue,
		 * disable the ack timer and clear the retransmit flag.
		 */

		skb = __skb_dequeue(&masc->ackw_queue);
		if (skb) {
			masc_skb_cb_t *masc_cb = (masc_skb_cb_t *) skb->cb;
			if (skb->protocol & MASC_INFO) {
				if (masc_cb->info_seqno == masc->info_seqno_ioctl) {
					switch (skb->protocol) {
					case MASC_TXOFF:
						masc->flg_xmit_on = 0;
						masc->flg_xmit_off = 1;
						masc->flg_ioctl_timeout = 0;
						wake_up_interruptible(&masc->ioctl_wait);
						break;
					case MASC_TXON:
						masc->flg_xmit_on = 1;
						masc->flg_xmit_off = 0;
						masc->flg_ioctl_timeout = 0;
						wake_up_interruptible(&masc->ioctl_wait);
						break;
					case MASC_SETNETID:
						masc->flg_ioctl_timeout = 0;
						wake_up_interruptible(&masc->ioctl_wait);
						break;
					default:
						break;
					}
				}
				masc_free_buf(masc, skb);
			} else {
				/*
				 * timestamp used for cleaning out stale entries
				 * from FH queue
				 */
				masc_cb->timestamp = jiffies;
				__skb_queue_tail(&masc->fh_queue, skb); 
			}
		} else {
			printk(KERN_ERR MNAME "%s: received unsolicited ack!\n", masc->if_name);
		}
		masc->flg_retran = 0;
		masc->flg_initack = 0;
		del_timer(&masc->ack_timer);

		MASC_DBG_PRINTK((KERN_INFO MNAME "ack processed\n"));
	}

	masc_kick_tty(masc);
}

/*
 * Process a NACK frame from the modem by retransmitting the
 * current ``ack packet''. We don't actually retransmit here,
 * but we set a flag indicating that a retransmit is needed.
 * This is detected in masc_fill_transmit_buf which is called
 * by masc_kick_tty. Device is locked.
 */

static void masc_process_nack(masc_dev_t *masc, unsigned char *pak, int len)
{
	if (len != 3)
		return;

	if (++masc->nack_count < 30) {
		masc->flg_racks = 0;
		masc->flg_retran = 1;
		MASC_DBG_PRINTK((KERN_INFO MNAME "nack resulting in retransmit\n"));
		masc_kick_tty(masc);
	} else {
		struct sk_buff *skb;

		printk(KERN_ERR MNAME "%s: modem bug: modem keeps rejecting correct frame!\n", masc->if_name);
		skb = __skb_dequeue(&masc->ackw_queue);
		if (skb) {
			masc_skb_cb_t *masc_cb = (masc_skb_cb_t *) skb->cb;
			if ((skb->protocol & MASC_INFO) && masc_cb->info_seqno == masc->info_seqno_ioctl)
				wake_up_interruptible(&masc->ioctl_wait);
			masc_free_buf(masc, skb);
		}
	}
}

/*
 * Search the queue of outstanding MASC packets (ones that not been sent over
 * the radio yet) and find one that matches the given sequence number,
 * otherwise return null. We disable interrupts while searching the
 * queue, and use unsafe operations. The queue is never longer than
 * MASC_MAX_FRAME, which is a small number---typically 2.
 * Instead of dequeuing packets and placing them to the
 * rear of the queue, how about looping through skb->next here? 
 * This would, however, violate the opaque sk_buff datatype,
 * and the kernel doesn't provide a skb_next() operation.
 * Device must be locked on entry.
 */

static struct sk_buff *masc_get_untransmitted(masc_dev_t *masc, int seqno)
{
	int qlen;

	qlen = skb_queue_len(&masc->fh_queue);
	while (qlen--) {
		struct sk_buff *skb = __skb_dequeue(&masc->fh_queue);
		if (skb->data[7] == seqno)
			return skb;
		__skb_queue_tail(&masc->fh_queue, skb);
	}

	return NULL;
}

/*
 * Send an info frame down the info_queue. A complete MASC frame
 * with correct checksum must be supplied. M frames should not
 * be sent this way; they belong to the mpak_queue.
 * Device locked.
 */

static void masc_send_info(masc_dev_t *masc, char *frame, int len, int type)
{
	struct sk_buff *skb = masc_alloc_buf(masc, MASC_SMALL, GFP_ATOMIC);
	masc_skb_cb_t *masc_cb;

	if (!skb) {
		printk(KERN_ERR MNAME "%s: unable to send info frame due to insufficient memory!\n", masc->if_name);
		return;
	}

	masc_cb = (masc_skb_cb_t *) skb->cb;
	memcpy(skb->data, frame, len);
	__skb_put(skb, len);
	skb->protocol = type;
	masc_cb->info_seqno = masc->info_seqno_gen++;
	__skb_queue_tail(&masc->info_queue, skb);
}

/*
 * Compute MASC checksum and send resulting info frame. A complete
 * MASC frame with uninitialized space reserved for the checksum
 * must be supplied. Device is locked.
 */

static void masc_cksum_send_info(masc_dev_t *masc, char *frame, int len, int type)
{
    int i, cksum = 0;

    for (i = 0; i < len - 3; i++)
	cksum ^= frame[i];

    frame[i++] = ((cksum >> 4) & 0xf)["0123456789ABCDEF"];
    frame[i] = (cksum & 0xf)["0123456789ABCDEF"];

    masc_send_info(masc, frame, len, type);
}

/*
 * Send F O frame to modem and wait for shutdown or timeout.
 * Called in ordinary process context, device not locked.
 */

static int masc_modem_off(masc_dev_t *masc)
{
	struct wait_queue wait = { current, NULL };
	int err = 0;

	masc_lock(masc);

	/*
	 * It's already off, bail!
	 */

	if (masc->flg_modem_off)
		goto bail;

	/*
	 * If someone is already turning off, just join the wait, otherwise
	 * initiate the turn off.
	 */

	if (!masc->flg_modem_off_w) {
		masc->flg_modem_off_w = 1;
		masc_send_info(masc, masc_turnoff_frame, sizeof masc_turnoff_frame, MASC_INFO);
		masc_kick_tty(masc);
	}

	/*
	 * We use our own timer rather than schedule_timeout here because we
	 * want to extend the timeout as the modem is shutting down if it has
	 * queued packets to bounce back to us, and does so in a timely
	 * fashion.
	 */

	masc->off_timer.expires = jiffies + MASC_TURNOFF_TIMEOUT*HZ;
	add_timer(&masc->off_timer);

	add_wait_queue(&masc->off_wait, &wait);

	current->state = TASK_INTERRUPTIBLE;

	if (!masc->flg_modem_off) {
		MOD_INC_USE_COUNT;	/* don't want driver to disappear while we are waiting! */
		masc_unlock(masc);
		schedule();
		masc_lock(masc);
		MOD_DEC_USE_COUNT;
		if (signal_pending(current)) {
			MASC_DBG_PRINTK((KERN_INFO MNAME "modem turn off wait cancelled by signal\n"));
			err = -EINTR;
		}
	}

	current->state = TASK_RUNNING;

	remove_wait_queue(&masc->off_wait, &wait);

	if (!err && !masc->flg_modem_off) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "modem turn off wait cancelled by timeout!\n"));
		err = -ETIME;
	}

	del_timer(&masc->off_timer);

bail:
	masc_unlock(masc);

	return err;
}

/*
 * Code to begin blocking ioctl.
 * Called with device locked.
 */

static int masc_begin_ioctl(masc_dev_t *masc, struct wait_queue *wait)
{
	/*
	 * Someone already doing ioctl. Report error.
	 */
	if (masc->flg_ioctl_busy)
		return -EBUSY;
	masc->flg_ioctl_busy = 1;
	masc->flg_ioctl_timeout = 1;
	masc->info_seqno_ioctl = masc->info_seqno_gen;
	add_wait_queue(&masc->ioctl_wait, wait);
	current->state = TASK_INTERRUPTIBLE;
	return 0;
}

/*
 * Do the ioctl wait.
 * Called with device locked.
 */

static int masc_ioctl_wait(masc_dev_t *masc, struct wait_queue *wait)
{
	MOD_INC_USE_COUNT;
	masc_unlock(masc);
	schedule_timeout(MASC_IOCTL_TIMEOUT * HZ);
	masc_lock(masc);
	MOD_DEC_USE_COUNT;

	if (signal_pending(current)) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "modem ioctl wait cancelled by signal\n"));
		return -EINTR;
	}

	if (masc->flg_ioctl_timeout)
		return -ETIME;

	return 0;
}

/*
 * Code to finalize blocking ioctl.
 * Called with device locked.
 */

static void masc_end_ioctl(masc_dev_t *masc, struct wait_queue *wait)
{
	current->state = TASK_RUNNING;
	remove_wait_queue(&masc->ioctl_wait, wait);
	masc->flg_ioctl_busy = 0;
}

/*
 * Turn modem's transmitter off.
 */

static int masc_transmitter_off(masc_dev_t *masc)
{
	struct wait_queue wait = { current, NULL };
	int err = 0;

	masc_lock(masc);

	/*
	 * If an ioctl is in progress, or the transmitter
	 * is known already to be off, bail now.
	 */

	if ((err = masc_begin_ioctl(masc, &wait)) == 0 && !masc->flg_xmit_off)  {
		masc_send_info(masc, masc_xmit_off_frame, sizeof masc_xmit_off_frame, MASC_TXOFF);
		masc_kick_tty(masc);
		err = masc_ioctl_wait(masc, &wait);
	}
	
	masc_end_ioctl(masc, &wait);
	masc_unlock(masc);

	return err;
}

static int masc_transmitter_on(masc_dev_t *masc)
{
	struct wait_queue wait = { current, NULL };
	int err = 0;

	masc_lock(masc);

	if ((err = masc_begin_ioctl(masc, &wait)) == 0 && !masc->flg_xmit_on) {
		masc_send_info(masc, masc_xmit_on_frame, sizeof masc_xmit_on_frame, MASC_TXON);
		masc_kick_tty(masc);
		err = masc_ioctl_wait(masc, &wait);
	}

	masc_end_ioctl(masc, &wait);
	masc_unlock(masc);

	return err;
}

static int masc_set_netid(masc_dev_t *masc, unsigned long id)
{
	struct wait_queue wait = { current, NULL };
	int err = 0;

	masc_lock(masc);

	if ((err = masc_begin_ioctl(masc, &wait)) == 0) {
		char set_netid_frame[22];
		sprintf(set_netid_frame, "^0015F R%04X,%04X:__\r",
			(unsigned) ((id >> 16) & 0xffff), (unsigned) (id & 0xffff));
		masc_cksum_send_info(masc, set_netid_frame, 21, MASC_SETNETID);
		masc_kick_tty(masc);
		err = masc_ioctl_wait(masc, &wait);
	}

	masc_end_ioctl(masc, &wait);
	masc_unlock(masc);
	return 0;
}

static int masc_wait_drain(masc_dev_t *masc)
{
	struct wait_queue wait = { current, NULL };
	int err = 0;

	masc_lock(masc);

	if (skb_queue_len(&masc->mpak_queue) > 0)
		goto wait;
	if (skb_queue_len(&masc->ctrl_queue) > 0)
		goto wait;
	if (skb_queue_len(&masc->info_queue) > 0)
		goto wait;

	goto done;
wait:
	add_wait_queue(&masc->drain_wait, &wait);
	current->state = TASK_INTERRUPTIBLE;
	MOD_INC_USE_COUNT;
	masc_unlock(masc);
	if (schedule_timeout(MASC_IOCTL_TIMEOUT * HZ) == 0)
		err = -ETIME;
	else if (signal_pending(current))
		err = -EINTR;
	masc_lock(masc);
	MOD_DEC_USE_COUNT;
done:
	masc_unlock(masc);
	return err;
}

/*
 * Thunk an MPAK header to an ethernet header. This is done
 * for packets received with Mobitex HPID 5 (the packet driver
 * traffic protocol ID). See the comment for masc_dev_hard_header.
 * The MPAK is assumed to have a payload of at least two bytes;
 * the first two bytes encode the protocol type.
 */

static void masc_mpak_to_ether(struct sk_buff *skb)
{
	unsigned char eh[14];
	unsigned char *mpak = skb->data;

	memset(eh, 0, sizeof eh);
	memcpy(eh, mpak + 3, 3);
	memcpy(eh + 6, mpak, 3);
	memcpy(eh + 12, mpak + 12, 2);
	memcpy(mpak, eh, 14);
}

/* 
 * Convert an ethernet header to an MPAK header. This is done
 * for all packets that have a protocol ID other than ETH_P_ALL
 * or ETH_P_MOBITEX. I.e. anything that was passed through
 * masc_dev_hard_header and got a fake ethernet header.
 * Note that the type field actually stays put under the
 * transformation, but it's clearer to not depend on that.
 */

static void masc_ether_to_mpak(struct sk_buff *skb)
{
	unsigned char mh[14];
	unsigned char *eth = skb->data;

	/*
	 * Thunk addresses and type field
	 */

	memcpy(mh, eth + 6, 3);
	memcpy(mh + 3, eth, 3);
	memcpy(mh + 12, eth + 12, 2);

	/*
	 * Set the Mobitex subscription flags and traffic state 
	 * byte to the appropriate values.
	 */

	mh[6] = MBTX_OUTGOING_FLAGS;
	mh[7] = MBTX_OUTGOING_TYPE;
	mh[8] = 0;	/* time field, we just zero it out */
	mh[9] = 0;
	mh[10] = 0;
	mh[11] = MBTX_PACKET;	/* packet driver HPID */

	memcpy(eth, mh, 14);
}


/*
 * Allocate a sk_buff and decapsulate the given MASC M frame into
 * it. Device is locked.
 */

static struct sk_buff *masc_mframe_decap(masc_dev_t *masc, unsigned char *pak, int len)
{
	struct sk_buff *skb = masc_alloc_buf(masc, MASC_BIG, GFP_ATOMIC);
	int count = 0;
	unsigned char *out;

	if (!skb)
		return NULL;

	pak += 7;	/* skip to the MASC data field  ^0000M:<data> 	*/
	len -= 7;	/*                                ^ seven chars	*/

	out = skb->data;

	/*
	 * Extract pairs of hex digits, but ignore the last
	 * three characters (the two digit checksum and carriage return).
	 * Convert the pairs into bytes. If an invalid hex digit
	 * is detected, free the buffer and bail out.
	 */

	while (len > 3) {
		int ch = masc_read_hex(pak, 2);
		if (ch == -1) {
			masc_free_buf(masc, skb);
			return NULL;
		}
		*out++ = ch;
		count++;
		pak += 2;
		len -= 2;
	}

	__skb_put(skb, count);
	return skb;
}

/*
 * Process received M frame. If it is a HPDATA type M frame whose protocol
 * ID (HPID) matches one of our three supported protocol types, we pass
 * it up to the networking layer. Also, if a packet is returned by the
 * network, we simply drop it. Device is locked.
 */

static void masc_process_mframe(masc_dev_t *masc, unsigned char *pak, int len)
{
	int tfc_state, class, type;
	unsigned char *mpak;
	struct sk_buff *skb;
	
	if ((skb = masc_mframe_decap(masc, pak, len)) == NULL) {
		printk(KERN_ERR MNAME "%s: dropping received packet due to insufficient memory\n", masc->if_name);
		masc->net_stats.rx_dropped++;
		return;
	}

	mpak = skb->data;

	tfc_state = (mpak[6] & MBTX_TFC_MASK) >> MBTX_TFC_SHIFT;
	class = (mpak[7] & MBTX_CLASS_MASK) >> MBTX_CLASS_SHIFT;
	type = (mpak[7] & MBTX_TYPE_MASK) >> MBTX_TYPE_SHIFT;

	/*
	 * The default protocol is ETH_P_MOBITEX, which marks a Mobitex packet
	 * that we have not otherwise recognized as an IP, ARP or RARP packet.
	 * Applications can install taps for this packet type to do raw Mobitex
	 * communications.
	 */

	skb->protocol = htons(ETH_P_MOBITEX);
	skb->dev = masc->dev;
	skb->mac.raw = skb->data;

	switch (tfc_state) {
	case MBTX_FROM_MAIL:
		masc->rx_mailbox++;
		/* fall through */
	case MBTX_OK:
		switch (class) {
		case MBTX_PSUBCOM:
			switch (type) {
			case MBTX_HPDATA:
				/*
				 * For packet driver packets, we extract
				 * the protocol id, and do the conversion
				 * to ethernet header format. Other packet
				 * types get left alone.
				 */
				if (mpak[11] == MBTX_PACKET && skb->len >= 14) {
					skb->protocol = htons(((unsigned) mpak[12] & 0xff) << 8 | (mpak[13] & 0xff));
					masc_mpak_to_ether(skb);
					skb_pull(skb, ETH_HLEN); 
				}
				break;
			default:
				break;
			}
			break;
		case MBTX_DTESERV:
			switch (type) {
			case MBTX_GROUPLIST:
				/*
				 * We have received a group list. If there are
				 * at least two addresses in it, take the second
				 * one and install it as a hardware broadcast
				 * address. The first address is the ``All
				 * Terminals Group'' MAN, which cannot be used
				 * for broadcasts in most Mobitex networks. The
				 * second address is most likely a group
				 * representing all of the modems belonging to
				 * a particular subscriber, so this is a more
				 * appropriate choice.  Of course, if the user
				 * has overriden the broadcast address using
				 * the masc_bcast[] array, then don't install
				 * the modem's address.
				 */
				MASC_DBG_PRINTK((KERN_INFO MNAME "group list received\n"));
				if (mpak[8] > 1 && masc_bcast[masc->line] == 0) {
					struct device *dev = masc->dev;
					memset(dev->broadcast, 0, ETH_ALEN);
					dev->broadcast[0] = mpak[12];
					dev->broadcast[1] = mpak[13];
					dev->broadcast[2] = mpak[14];
				}
				break;
			default:
				break;
			}
			break;
		default:
			break;
		}
		break;
	case MBTX_IN_MAIL:
		masc->tx_mailbox++;
		break;
	case MBTX_NO_TRANSFER:
		masc->rx_notran++;
		break;
	case MBTX_ILLEGAL:
		masc->rx_illegal++;
		break;
	case MBTX_CONGEST:
		masc->rx_congest++;
		break;
	case MBTX_ERROR:
		masc->rx_errcount++;
		break;
	case MBTX_BUSY:
		masc->rx_busy++;
		break;
	}

	MASC_DBG_PRINTK((KERN_INFO MNAME "MPAK protocol is %x\n", (unsigned) ntohs(skb->protocol)));
	masc_unlock(masc);
	netif_rx(skb);
	masc_lock(masc);
	masc->net_stats.rx_packets++;
}

/*
 * Handle error code from a F K, R or N frame. The packet sequence number is
 * specified as -1 if the error does not correspond to any packet.
 * Device is locked.
 */

static void masc_handle_error(masc_dev_t *masc, int code, int seq)
{
	struct sk_buff *skb = NULL;
	enum { kick, drop, requeue } action = kick;

	if (seq != -1)
		skb = masc_get_untransmitted(masc, seq);

	if (code != -1) {
		masc->errhist[0] = masc->errhist[1];
		masc->errhist[1] = masc->errhist[2];
		masc->errhist[2] = masc->errhist[3];
		masc->errhist[3] = code;
	}

	switch (code) {
	case -1:
		action = drop;
		break;
	case 0x01:	/* modem received order to stop traffic		*/
		masc->flg_tstop = 1;
		break;
	case 0x02:	/* modem received order to start traffic	*/
		masc->flg_tstop = 0;
		action = kick;
		break;
	case 0x10:	/* MPAK returned due to network stop order	*/
		masc->flg_tstop = 1;
		action = requeue;
		break;
	case 0x13:	/* MPAK returned---transmission buffer full	*/
		printk(KERN_ERR MNAME "%s: radio modem bounced packet due to buffer full!\n",
			masc->if_name);
		action = requeue;
		break;
	case 0x18:	/* The originating MAN is unrecognized		*/
		printk(KERN_ERR MNAME "%s: radio modem bounced packet due to wrong originating address!\n",
			masc->if_name);
		action = drop;
		break;
	case 0x51:	/* Invalid MPAK type, must not be transmitted	*/
		action = drop;
		break;
	case 0x52:	/* Invalid MPAK state				*/
	case 0x53:	/* Invalid MPAK flags				*/
	case 0x54:	/* Invalid send list				*/
	case 0x55:	/* Invalid MPAK length				*/
	case 0x56:	/* Invalid addressee in MPAK			*/
		action = drop;
		break;
	case 0x62:	/* MPAK returned on request (F I or F O)	*/
		action = drop;
		/* restart the turn-off timeout to give modem more time	*/
		del_timer(&masc->off_timer);
		masc->off_timer.expires = jiffies + MASC_TURNOFF_TIMEOUT*HZ;
		add_timer(&masc->off_timer);
		break;
	case 0x66:	/* Invalid network ID				*/
		printk(KERN_ERR MNAME "%s: radio modem rejected attempt to set invalid network id!\n",
			masc->if_name);
		break;
	case 0xA0:	/* Transmission buffer full			*/
	case 0xE0:	/* Radio transmission has failed		*/
	case 0xF0:	/* Transmit queue full; max number of MPAKS	*/
	case 0xFD:	/* Other error; modem not able to transmit	*/
		action = requeue;
		break;

	/*
	 * These errors we either don't expect, cannot do anything
	 * about, or they are just informative messages that require
	 * no action.
	 */
	case 0x0A:	/* reception buffer full			*/
	case 0x0B:	/* reception buffers free			*/
	case 0x16:	/* Login denied, MAN already in flexlist	*/
	case 0x61:	/* TEMP_DEFAULT_LIST received incorrectly	*/
	case 0x63:	/* User changes state into already active mode	*/
	case 0x65:	/* Mode change already activated		*/
	case 0x67:	/* Mode MPAK with invalid parameter		*/
	case 0xF8:	/* MPAK returned because transmitter is off	*/
	default:	/* unknown or garbled */
		action = drop;	/* just in case, so we can keep going	*/
		break;
	}

	if (skb) {
		switch (action) {
		case drop:
			masc_free_buf(masc, skb);
			masc->net_stats.tx_dropped++;
			break;
		case requeue:
			__skb_queue_tail(&masc->mpak_queue, skb);
			masc->net_stats.tx_errors++;
			break;
		case kick:
			__skb_queue_tail(&masc->fh_queue, skb);
			break;
		}
		masc_kick_tty(masc);
	}
}

static void masc_send_init(masc_dev_t *masc)
{

	/*
	 * Send INIT frame to the modem in response.
	 */

	masc_send_info(masc, masc_init_frame, sizeof masc_init_frame, MASC_INIT);
	masc->flg_initsent = 1;

	/*
	 * Send request to obtain modem's MAN address if we don't
	 * already have it.
	 */

	if (!masc->flg_haveman)
		masc_send_info(masc, masc_getman_frame, sizeof masc_getman_frame, MASC_INFO);

	/*
	 * We start the battery and signal strength interrogation timer.
	 * Give it just one second for the first sample. Subsequent samples
	 * will be taken every MASC_BATT_POLL seconds.
	 */

	del_timer(&masc->bat_timer);
	masc->bat_timer.expires = jiffies + 1*HZ;
	add_timer(&masc->bat_timer);
}

/*
 * Process received information frame based the type specified in its
 * text field. A wide variety of information is communicated through
 * info frames, in addition to MPAKS. We perform additional sanity
 * checking on the information frames as opportunity arises, and
 * return 0 so that the caller can NACK a badly received packet.
 * Device is locked.
 */

static int masc_process_infoframe(masc_dev_t *masc, unsigned char *pak, int len)
{
	int seqno, err, value;
	long man;
	struct sk_buff *skb;

	switch (pak[5]) {
	case 'B':		/* INIT frame		*/
		masc_send_init(masc);
		break;
	case 'M':		/* MPAK-carrying frame	*/
		masc_process_mframe(masc, pak, len);
		break;
	case 'F':
		if (pak[6] != ' ')
			return 0;
		switch (pak[7]) {
		case 'H':	/* FH frame---MPAK radioed to base	*/
			MASC_DBG_PRINTK((KERN_INFO MNAME "FH frame (%c) being processed\n", (int) pak[8]));
			seqno = pak[8];
			if (!isdigit(seqno))
				return 0;
			skb = masc_get_untransmitted(masc, seqno);
			if (skb) {
				masc_free_buf(masc, skb);
				masc->net_stats.tx_packets++;
			} else {
				printk(KERN_INFO "%s: received FH frame for packet it didn't send\n",
					masc->if_name);
			}
			masc->traf_hist[(masc->traf_curr + MASC_MAX_AVG_AIRTIME) % (sizeof masc->traf_hist / sizeof masc->traf_hist[0])]++;
			masc->traf_count++;
			/* if we are getting FH's we must be in coverage */
			masc->flg_nocontact = 0;
			break;
		case 'K':	/* FK frame---error condition		*/
			err = masc_read_hex(pak + 8, 2);
			masc_handle_error(masc, err, -1);
			break;	
		case 'G':	/* Lost contact with Mobitex network	*/
			masc->flg_nocontact = 1;
			/* assume signal strength is zero */
			masc->sign = 0;
			break;
		case 'F':	/* In contact with Mobitex network 	*/
			masc->flg_nocontact = 0;
			/* find out what the signal strength is ASAP */
			if (skb_queue_len(&masc->info_queue) < 3)
				masc_send_info(masc, masc_getsign_frame, sizeof masc_getsign_frame, MASC_INFO);
			break;
		case 'O':	/* Modem is about to turn off		*/
			del_timer(&masc->off_timer);
			/* was turnoff called for in this session? */
			if (!masc->flg_modem_off_w) {
				/* TODO: take steps to turn modem back on! */
			}
			masc->flg_modem_off_w = 0;
			masc->flg_modem_off = 1;
			wake_up_interruptible(&masc->off_wait);
			break;
		case 'P':	/* Modem is telling us the MAN number!	*/
			if (len != 18)
				return 0;
			man = masc_read_hex(pak + 8, 6);	
			if (man != -1 && man != 0 && man != 1) {
				unsigned char *addr = masc->dev->dev_addr;
				memset(addr, 0, ETH_ALEN);
				*addr++ = (man >> 16) & 0xff;
				*addr++ = (man >> 8) & 0xff;
				*addr++ = man & 0xff;
				masc->flg_haveman = 1;
			}
			break;
		case 'A':	/* Battery, RSSI or power save info	*/
			if (len != 17)
				return 0;
			if (pak[10] != ',')
				return 0;
			switch (masc_read_hex(pak + 8, 2)) {
			case -1:	/* bad hex */
				return 0;
			case 1:	/* power saving mode info	*/
				break;	/* not supported yet	*/
			case 2:	/* RSSI value			*/
				value = masc_read_hex(pak + 11, 2);
				if (value == -1)
					return 0;
				masc->sign = value;
				break;
			case 3: /* Battery value		*/
				value = masc_read_hex(pak + 11, 2);
				if (value == -1)
					return 0;
				masc->batt = value;
				break;
			}
			break;
		default:	/* Unsupported F frame			*/
			break;
		}
		break;
	case 'N':		/* Return of untransmitted MPAK		*/
	case 'R':		/* Return of errnoeous MPAK		*/
		if (pak[6] == '\r')
			return 1;
		if (pak[6] != ' ')
			return 0;
		if (pak[7] == ',') {
			err = -1;
			seqno = pak[8];
		} else {
			if (pak[9] != ',')
				return 0;
			err = masc_read_hex(pak + 7, 2);
			seqno = pak[10];
		}
		masc_handle_error(masc, err, seqno);
		break;	
	default:		/* drop unrecognized info frame		*/
		break;
	}
	return 1;
}

/*
 * Process frame received from modem as a MASC packet, taking
 * into account the overflow and toss (parity error) flag.
 * Device is locked.
 */

static void masc_process_frame(masc_dev_t *masc)
{
	unsigned char *rx_buf = masc->rx_buf;
	int rx_count = masc->rx_count;
	int overflow = masc->flg_oflow;
	int toss = masc->flg_toss;

	MASC_DBG_PRINTK((KERN_INFO MNAME "processing received frame, length %x\n", (unsigned) rx_count));
	rx_buf[rx_count] = 0;
	MASC_DBG_PRINTK((KERN_INFO MNAME "frame (truncated to 64 chars): %.64s\n", rx_buf));

	masc->flg_oflow = 0;
	masc->flg_toss = 0;
	masc->rx_count = 0;

	if (overflow)
		return;

	if (rx_count < 2)
		return;

	if (rx_buf[0] != '^' && rx_buf[rx_count-1] != '\r')
		return;

	if (isxdigit(rx_buf[1])) {	/* information or init frame	*/
		if (masc_infoframe_basic_validate(rx_buf, rx_count)) {
			if (!toss && masc_infoframe_validate(rx_buf, rx_count)
					&& masc_process_infoframe(masc, rx_buf, rx_count)) {
				masc->send_ack += 1;
				masc->send_ack &= 1;
				masc_send_ack(masc, masc->send_ack); 
			} else {
				masc_send_ctrl(masc, "?");	/* send NACK */
			}
			masc_kick_tty(masc);
		}
	} else {			/* control or garbled frame	*/
		if (toss)
			return;
		switch (rx_count) {
		case 4:
			if (rx_buf[1] == '*')	/* ACK frame */
				masc_process_ack(masc, rx_buf, rx_count);
			break;
		case 3:
			switch (rx_buf[1]) {
			case '?':		/* NACK frame	*/
				masc_process_nack(masc, rx_buf, rx_count);
				break;
			case '!':		/* RACK frame	*/
				masc_send_ack(masc, masc->send_ack);
				masc_kick_tty(masc);
				break;
			case '#':		/* SENS frame	*/
				masc_send_ctrl(masc, "&");
				masc_kick_tty(masc);
				break;
			case '&':		/* SACK frame	*/
				break;
			default:		/* whatever	*/
				break;
			}
		default:
			break;
		}
	}
}

/*
 * Free the buffers on all the MASC queues.
 * This is performed when the device is shut down.
 * At this point, all timers, interrupts and other bottom-half callbacks
 * have stopped and we are disconnected from the networking layer.
 * So it's safe to use non-atomic queue operations even with
 * device unlocked.
 */

static void masc_free_buffers(masc_dev_t *masc)
{
	struct sk_buff *skb;

	while ((skb = __skb_dequeue(&masc->mpak_queue)) != NULL)
		kfree_skb(skb);

	while ((skb = __skb_dequeue(&masc->info_queue)) != NULL)
		kfree_skb(skb);

	while ((skb = __skb_dequeue(&masc->ctrl_queue)) != NULL)
		kfree_skb(skb);

	while ((skb = __skb_dequeue(&masc->free_queue[MASC_SMALL])) != NULL)
		kfree_skb(skb);

	while ((skb = __skb_dequeue(&masc->free_queue[MASC_BIG])) != NULL)
		kfree_skb(skb);

	while ((skb = __skb_dequeue(&masc->ackw_queue)) != NULL)
		kfree_skb(skb);

	while ((skb = __skb_dequeue(&masc->fh_queue)) != NULL)
		kfree_skb(skb);
}

/*
 * Network device operations.
 *
 */

/*
 * Function called by the networking layer to indicate that the network
 * device is going up! We don't have to do anything here, but we perform
 * a simple sanity check to ensure we are connected to a tty.
 */

static int masc_dev_open(struct device *dev)
{
	masc_dev_t *masc;
	struct tty_struct *tty;

	if (!(masc = dev->priv)) {
		printk(KERN_ERR MNAME "null pointer to masc structure in net device!\n");
		return -ENXIO;
	}

	if (!(tty = masc->tty)) {
		printk(KERN_ERR MNAME "interface wants to go up, but tty is unavailable!\n");
		return -ENXIO;
	}

	return 0;
}

static int masc_dev_ioctl(struct device *dev, struct ifreq *ifr, int cmd)
{
	return -EINVAL;	/* no ioctl's for now */
}

/*
 * Stop the device from transmitting
 */

static int masc_dev_stop(struct device *dev)
{
	masc_dev_t *masc = dev->priv;

	if (masc->tty == NULL)
		return -EBUSY;

	/*
	 * Make it appear as though the device is in a
	 * busy state, and prevent tty wakeups.
	 */

	masc_lock(masc);
	masc->tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
	masc_unlock(masc);
	dev->tbusy = 1;

	return 0;
}

/*
 * Take a raw packet and encapsulate it as a MASC info frame. Device is locked.
 */


static void masc_encaps_mframe(masc_dev_t *masc, struct sk_buff *sk_in, struct sk_buff *sk_out)
{
	unsigned char *in = sk_in->data, *out = sk_out->data;
	int len = 1 + 4 + 3 + 1 + sk_in->len * 2 + 2 + 1;
	int cksum, count = sk_in->len;
	int proto = ntohs(sk_in->protocol);
	/*
	 * IP and other protocol packets have been built with ethernet headers. 
	 * We undo the emulation and switch to MPAK since we are entering
	 * Mobitex world. ETH_P_MOBITEX packet are assumed to be in MPAK
	 * format already, and ETH_P_ALL is an escape hatch to put arbitrary
	 * bits out, so we leave those two types alone.
	 */

	if (proto != ETH_P_MOBITEX && proto != ETH_P_ALL)
		masc_ether_to_mpak(sk_in);

	cksum = *out++ = '^';

	sprintf(out, "%04X", len);
	cksum ^= *out++;
	cksum ^= *out++;
	cksum ^= *out++;
	cksum ^= *out++;

	cksum ^= *out++ = 'M';
	cksum ^= *out++ = ' ';
	cksum ^= *out++ = '0' + masc->seqno;
	cksum ^= *out++ = ':';

	while (count--) {
		cksum ^= *out++ = ((*in >> 4) & 0xf)["0123456789ABCDEF"];
		cksum ^= *out++ = (*in++ & 0xf)["0123456789ABCDEF"];
	}

	*out++ = ((cksum >> 4) & 0xf)["0123456789ABCDEF"];
	*out++ = (cksum & 0xf)["0123456789ABCDEF"];

	*out++ = '\r';

	__skb_put(sk_out, out - sk_out->data);

	if (masc->seqno++ == 9)
		masc->seqno = 0;

	sk_out->protocol = MASC_MPAK;
}

/*
 * Called from the networking system to pass a packet down to us, which
 * we turn into an M frame and ship it to the modem.
 */

static int masc_dev_xmit(struct sk_buff *skb, struct device *dev)
{
	masc_dev_t *masc = dev->priv;
	struct sk_buff *masc_skb;

	masc_lock(masc);

	MASC_DBG_PRINTK((KERN_INFO MNAME "masc_dev_xmit on %s\n", masc->if_name));

	/*
	 * We check the size of our M frame queue. If it is too long,
	 * we tell the networking system to hold its horses. Note that
	 * we queue M frames even though we might not have done the INIT/INIT
	 * handshake with the modem. The masc_fill_transmit_buf function won't
	 * send any frames to the modem until INIT is received from the modem.
	 * The low priority of the M frame queue will then ensure that our INIT
	 * will get through ahead of any queued M frames.
	 */

	if (skb_queue_len(&masc->mpak_queue) >= MASC_MAX_QUEUE) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "masc_dev_xmit on %s balked: too many packets queued\n", masc->if_name));
		dev->tbusy = 1;		/* indicate that device can't take any more packets */
		masc_kick_tty(masc);	/* kick-start downward flow in case it's stopped */
		masc_unlock(masc);
		return 1;
	}

	/*
	 * Otherwise we try to allocate a new MASC packet buffer.
	 * If we can't allocate packet, we simply balk.
	 */

	if ((masc_skb = masc_alloc_buf(masc, MASC_BIG, GFP_ATOMIC)) == NULL) {
		printk(KERN_ERR MNAME "%s: masc_dev_xmit dropping packet due to insufficient memory\n", masc->if_name);
		dev_kfree_skb(skb);	/* just toss the packet */
		masc_unlock(masc);
		return 0;
	}

	/*
	 * We take the raw MPAK and encapsulate it into an M frame MASC packet,
	 * giving an appropriate sequence number, etc.
	 * After this point, we have accepted the packet, and so it is
	 * our responsibility to free the sk_buff we have been given
	 * and to return a zero.  The header construction has already been done
	 * for this packet so it is a valid Mobitex PSUBCOM HPDATA packet.
	 */

	masc_encaps_mframe(masc, skb, masc_skb);


	MASC_DBG_PRINTK((KERN_INFO MNAME "%s: encapsulated frame\n", masc->if_name));

	/*
	 * We stuff the packet down into the queue designated for MPAK-carrying
	 * MASC frames (M frames).
	 */

	__skb_queue_tail(&masc->mpak_queue, masc_skb);

	/*
	 * We kick some data down to the tty if it is not already
	 * busy transmitting.
	 */

	masc_kick_tty(masc);
	masc_unlock(masc);

	dev_kfree_skb(skb);
	return 0;
}

/*
 * Gather traffic statistics into a standard form. One of the places
 * this information ends up is on the output of the ifconfig command,
 * allowing the user to view the statistics.
 */

static struct net_device_stats *masc_dev_stats(struct device *dev)
{
	masc_dev_t *masc = dev->priv;

	memset(&masc->net_stats_copy, 0, sizeof masc->net_stats_copy);
	masc_lock(masc);
	masc->net_stats_copy = masc->net_stats;
	masc_unlock(masc);

	return &masc->net_stats_copy;
}

/*
 * We are called upon by the network layer to build a hardware header.  What we
 * actually do is instead of building a Mobitex packet (MPAK) header, we build
 * an ethernet header. Why? Because ``all the world's an ethernet'' in Linux
 * networking. By emulating ethernet, we allow various packet capture tools
 * such as tcpdump to work with this driver without modification.  Otherwise
 * these tools would have to be modified to support the MPAK header.
 */

static int masc_dev_hard_header(struct sk_buff *skb, struct device *dev,
		unsigned short type, void *daddr,
		void *saddr, unsigned int len)
{
	MASC_DBG_PRINTK((KERN_INFO MNAME "masc_dev_hard_header called\n"));

	/*
	 * If it's already a Mobitex packet, don't do anything.
	 * Otherwise we would end up encapsulating a Mobitex packet
	 * in a Mobitex packet, which isn't the aim here.
	 * The type ETH_P_ALL is assumed to specify a packet whose
	 * bit contents we want to send onto the wire exactly, so
	 * we treat it the same way.
	 */

	if (type == ETH_P_MOBITEX || type == ETH_P_ALL) {
		return 0;
	} else {
		unsigned char *eth = skb_push(skb, ETH_HLEN);
		int type_field = (type == ETH_P_802_3) ? len : type;

		eth[12] = (type_field >> 8) & 0xff;
		eth[13] = type_field & 0xff;

		if (saddr != NULL)
			memcpy(eth + 6, saddr, ETH_ALEN);
		else
			memcpy(eth + 6, dev->dev_addr, ETH_ALEN);

		if (daddr != NULL) {
			memcpy(eth, daddr, ETH_ALEN);
			return dev->hard_header_len;
		}

		return -dev->hard_header_len;
	}
}

static int masc_dev_rebuild_header(struct sk_buff *skb)
{
	unsigned char *hdr = skb->data;

	MASC_DBG_PRINTK((KERN_INFO MNAME "masc_dev_rebuild_header called\n"));

#ifdef CONFIG_INET

	if (ntohs(skb->protocol) != ETH_P_IP) {
		printk(KERN_DEBUG MNAME "%s: cannot ARP non-IP addresses\n",
			skb->dev->name);
		return 1;
	}

	if (arp_find(hdr, skb))
		return 1;
#endif
	return 0;
}

static int masc_mac_addr(struct device *dev, void *ptr)
{
    struct sockaddr *addr = ptr;
    masc_dev_t *masc = dev->priv;
    masc_lock(masc);
    memcpy(dev->dev_addr, addr->sa_data, ETH_ALEN);
    masc_unlock(masc);
    return 0;
}


/*
 * ACK timeout
 */

static void masc_ack_timer(unsigned long arg)
{
	masc_dev_t *masc = (masc_dev_t *) arg;	/* ARGH! Ugly! Ugly! */

	masc_lock(masc);

	MASC_DBG_PRINTK((KERN_INFO MNAME "%s: ack timer\n", masc->if_name));

	del_timer(&masc->ack_timer);

	if (skb_queue_len(&masc->ackw_queue) > 0 && skb_queue_len(&masc->ctrl_queue) < 3) {
		masc_send_ctrl(masc, "!");		/* send RACK */
		masc->flg_racks = 1;			/* RACK sent flag */
		masc_kick_tty(masc);
		masc->ack_timer.expires = jiffies + MASC_ACK_TIMEOUT*HZ;
		add_timer(&masc->ack_timer);
	}

	masc_unlock(masc);
}

/*
 * Battery/RSSI timer
 */

static void masc_bat_timer(unsigned long arg)
{
	masc_dev_t *masc = (masc_dev_t *) arg;	/* Blech! */

	masc_lock(masc);

	MASC_DBG_PRINTK((KERN_INFO MNAME "%s: battery/RSSI timer\n", masc->if_name));

	del_timer(&masc->bat_timer);

	/*
	 * Guard against sending info frames before INIT because that is
	 * against MASC rules, and against filling the info queue if
	 * communication to the device isn't happening!
	 */

	if (masc->flg_initsent && skb_queue_len(&masc->info_queue) < 3) {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: sending battery status & signal strength request\n", masc->if_name));
		masc_send_info(masc, masc_getbatt_frame, sizeof masc_getbatt_frame, MASC_INFO);
		masc_send_info(masc, masc_getsign_frame, sizeof masc_getsign_frame, MASC_INFO);
		masc_kick_tty(masc);
	} else {
		MASC_DBG_PRINTK((KERN_INFO MNAME "%s: nothing to do\n", masc->if_name));
	}

	masc->bat_timer.expires = jiffies + MASC_BAT_POLL*HZ;
	add_timer(&masc->bat_timer);

	masc_unlock(masc);
}

static void masc_traf_timer(unsigned long arg)
{
	int traf_curr, do_kick = 0;
	masc_dev_t *masc = (masc_dev_t *) arg;	/* Blech! */

	masc_lock(masc);

	del_timer(&masc->traf_timer);
	traf_curr = masc->traf_curr + 1;
	if (traf_curr >= sizeof masc->traf_hist / sizeof masc->traf_hist[0])
		traf_curr = 0;
	if ((masc->traf_count -= masc->traf_hist[traf_curr]) < MASC_MAX_AIRED_PACKETS)
		do_kick = 1;
	masc->traf_hist[traf_curr] = 0;
	masc->traf_curr = traf_curr;
	if (do_kick)
		masc_kick_tty(masc);
	masc->traf_timer.expires = jiffies + 1*HZ;
	add_timer(&masc->traf_timer);

	masc_unlock(masc);
}

static void masc_off_timer(unsigned long arg)
{
	masc_dev_t *masc = (masc_dev_t *) arg;

	del_timer(&masc->off_timer);
	wake_up_interruptible(&masc->off_wait);
}

static void masc_init_timer(unsigned long arg)
{
	masc_dev_t *masc = (masc_dev_t *) arg;

	masc_lock(masc);
	del_timer(&masc->ini_timer);
	if (!masc->flg_initsent)
		masc_send_init(masc);
	masc_unlock(masc);
}

/*
 * Initialize the ``base class'' portion of a masc device. This
 * is called after the masc object is registered as a network
 * device.
 */

static int masc_dev_init(struct device *dev)
{
	masc_dev_t *masc = dev->priv;
	int num = dev->base_addr;

	/* sanity check */

	if (!masc || masc->magic != MASC_MAGIC) {
		printk(KERN_ERR MNAME "net device isn't a masc device!");
		return -ENODEV;
	}

	dev->hard_header = masc_dev_hard_header;
	dev->rebuild_header = masc_dev_rebuild_header;
	dev->set_mac_address = masc_mac_addr;
	dev->hard_start_xmit = masc_dev_xmit;
	dev->open = masc_dev_open;
	dev->stop = masc_dev_stop;
	dev->get_stats = masc_dev_stats;
	dev->do_ioctl = masc_dev_ioctl;

	dev->hard_header_len = ETH_HLEN;	/* ethernet header emulation */
	dev->mtu = MASC_MTU;
	dev->addr_len = ETH_ALEN;
	dev->tx_queue_len = MASC_MAX_DEV_QUEUE;
	dev->type = ARPHRD_ETHER;

	dev_init_buffers(dev);

	dev->flags = IFF_BROADCAST;

	/*
	 * Initialize the hardware broadcast device according
	 * to the the user-configurable values that can be
	 * specified on the insmod command line.
	 */

	if (num <= 9) {
		memset(dev->broadcast, 0, ETH_ALEN);
		dev->broadcast[0] = masc_bcast[num] >> 16 & 0xff;
		dev->broadcast[1] = masc_bcast[num] >> 8 & 0xff;
		dev->broadcast[2] = masc_bcast[num] & 0xff;
	}

	return 0;

}

/*
 * Find an existing, free masc device on the list. The global spinlock
 * is locked.
 */

static masc_dev_t *masc_find_free(void)
{
	masc_dev_t *masc = masc_list.next;

	while (masc != &masc_list) {
		int occup;
		masc_lock(masc);
		occup = masc->flg_occup;
		masc->flg_occup = 1;
		masc_unlock(masc);
		if (!occup)
			return masc;
		masc = masc->next;
	}

	return NULL;
}


/*
 * Allocate masc device using kmalloc(), initialize it, register it as a
 * network device and add it to the global masc device list.
 */

static masc_dev_t *masc_dev_alloc(void)
{
	static struct device blank_dev = { 0 };
	static masc_dev_t blank_masc = { MASC_MAGIC };
	masc_dev_t *masc;
	int masc_num;

	blank_masc.flg_occup = 1;

	masc_global_lock();

	if ((masc = masc_find_free()) != NULL) {
		masc_num = masc->line;
		blank_masc.next = masc->next;
		blank_masc.prev = masc->prev;
		*masc = blank_masc;
	} else {
		masc_global_unlock();
		masc = kmalloc(sizeof *masc, GFP_KERNEL);
		if (!masc) {
			printk(KERN_ERR MNAME "failure to allocate masc_dev!\n");
			goto failed_masc;
		}
		masc_global_lock();
		masc_num = masc_numdev++;
		*masc = blank_masc;
		masc->next = &masc_list;
		masc->prev = masc_list.prev;
		masc_list.prev->next = masc;
		masc_list.prev = masc;
	}
	masc_global_unlock();

	if (!(masc->tx_buf = kmalloc(MASC_BIG_BUF_SIZE, GFP_KERNEL))) {
		printk(KERN_ERR MNAME "failure to allocate tx buffer\n");
		goto failed_tx_buff;
	}

	if (!(masc->rx_buf = kmalloc(MASC_BIG_BUF_SIZE, GFP_KERNEL))) {
		printk(KERN_ERR MNAME "failure to allocate rx buffer\n");
		goto failed_rx_buff;
	}

	sprintf(masc->if_name, "masc%d", masc_num);

	if (dev_get(masc->if_name) != NULL) {
		printk(KERN_ERR MNAME "name %s is already in use!\n", masc->if_name);
		goto failed_name_grab;
	}

	masc->tx_head = masc->tx_buf;
	masc->line = masc_num;
	masc->send_ack = -1;
	masc->last_ack = -1;
	init_timer(&masc->ack_timer);
	masc->ack_timer.data = (unsigned long) masc;
	/* ARGH! The above cast is so lame, I want to scream!!! */
	masc->ack_timer.function = masc_ack_timer;	
	init_timer(&masc->bat_timer);
	masc->bat_timer.data = (unsigned long) masc;
	masc->bat_timer.function = masc_bat_timer;	
	init_timer(&masc->off_timer);
	masc->off_timer.data = (unsigned long) masc;
	masc->off_timer.function = masc_off_timer;	
	init_timer(&masc->traf_timer);
	masc->ini_timer.data = (unsigned long) masc;
	masc->ini_timer.function = masc_init_timer;	
	init_timer(&masc->ini_timer);
	masc->traf_timer.data = (unsigned long) masc;
	masc->traf_timer.function = masc_traf_timer;	
	masc->off_wait = NULL;
	masc->ioctl_wait = NULL;
	masc->drain_wait = NULL;
	masc->dev_base = blank_dev;
	masc->dev_base.name = masc->if_name;
	masc->dev_base.base_addr = masc_num;
	masc->dev_base.priv = masc;	
	masc->dev_base.next = NULL;
	masc->dev_base.init = masc_dev_init;
	masc->dev = &masc->dev_base;
	skb_queue_head_init(&masc->mpak_queue);
	skb_queue_head_init(&masc->info_queue);
	skb_queue_head_init(&masc->ctrl_queue);
	skb_queue_head_init(&masc->free_queue[MASC_SMALL]);
	skb_queue_head_init(&masc->free_queue[MASC_BIG]);
	skb_queue_head_init(&masc->ackw_queue);
	skb_queue_head_init(&masc->fh_queue);
	if (register_netdev(masc->dev) != 0) {
		printk(KERN_ERR MNAME "failed to register network device!\n");
		goto failed_register;
	}
	spin_lock_init(&masc->spinlock);
	return masc;

failed_register:
	kfree(masc->rx_buf);
failed_name_grab:
failed_rx_buff:
	kfree(masc->tx_buf);
failed_tx_buff:
	/* must not free masc here, it's been put on the list */
failed_masc:
	return NULL;
}

/*
 * Deallocate a masc structure
 */

static void masc_dev_dealloc(masc_dev_t *masc)
{
	struct tty_struct *tty = masc->tty;

	/*
	 * Break the tty --> masc link
	 */

	if (tty && tty->disc_data == masc)
		tty->disc_data = NULL;


	masc->tty = NULL;

	/*
	 * Kill timers!
	 */

	del_timer(&masc->ack_timer);
	del_timer(&masc->bat_timer);
	del_timer(&masc->off_timer);
	del_timer(&masc->ini_timer);
	del_timer(&masc->traf_timer);

	/*
	 * Wake up any processes that may still be waiting on this
	 * device.
	 */

	wake_up_interruptible(&masc->off_wait);
	wake_up_interruptible(&masc->ioctl_wait);
	wake_up_interruptible(&masc->drain_wait);

	/*
	 * Close net device, and unregister it
	 */

	dev_close(masc->dev);
	unregister_netdev(masc->dev);

	/*
	 * Free memory resources---all except for the masc_dev_t structure
	 * itself. This stays on the global list, and is reused.
	 */

	masc_free_buffers(masc);
	kfree(masc->rx_buf);
	kfree(masc->tx_buf);

	/*
	 * Setting this flag to zero marks this structure
	 * as free, so that the next masc_find_free()
	 * might be able to grab it.
	 */

	masc_global_lock();
	masc->flg_occup = 0;
	masc_global_unlock();
}

/*
 * The masc_tty_* functions form an implementation of the
 * abstract line discipline interface that is registered with the TTY
 * subsystem.
 */

/*
 * Called when the line discipline of a tty becomes MASC.
 */

static int masc_tty_open(struct tty_struct *tty)
{
	masc_dev_t *masc;

	if (tty->disc_data)
		return -EEXIST;

	if (!(masc = masc_dev_alloc()))
		return -ENFILE;

	masc->dev->tbusy = 1;
	masc->tty = tty;	
	tty->disc_data = masc;

	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);

	if (tty->driver.flush_buffer)
		tty->driver.flush_buffer(tty);

	MASC_DBG_PRINTK((KERN_INFO MNAME "allocated %s\n", masc->if_name));

	/*
	 * Traffic histogram timer is once every second over
	 * lifetime of driver.
	 */

	masc->traf_timer.expires = jiffies + 1*HZ;
	add_timer(&masc->traf_timer);

	/*
	 * Init timer sends INIT frame after a delay, unless
	 * it has been sent already in response to the modem's.
	 */

	masc->ini_timer.expires = jiffies + MASC_INIT_TIMEOUT*HZ;
	add_timer(&masc->ini_timer);

	MOD_INC_USE_COUNT;

	return masc->line;
}

/*
 * Called when the MASC line discipline is uninstalled
 * from the tty (because for instance the tty has closed).
 */

static void masc_tty_close(struct tty_struct *tty)
{
	masc_dev_t *masc = tty->disc_data;

	MASC_DBG_PRINTK((KERN_INFO MNAME "masc_tty_close called for %s\n", masc->if_name));

	if (!masc) {
		printk(KERN_ERR MNAME "tty struct disc_data pointer is null\n");
		return;
	}

	masc_dev_dealloc(masc);

	MASC_DBG_PRINTK((KERN_INFO MNAME "deallocated %s\n", masc->if_name));
	MOD_DEC_USE_COUNT;
}


#if 0

/*
 * These functions may be implemented in the future if some use
 * is found for them.
 */

static int masc_tty_read(struct tty_struct *tty, struct file *file,
		unsigned char *buf, unsigned int nr)
{
}

static int masc_tty_write(struct tty_struct *tty, struct file *file,
		const unsigned char *buf, unsigned int nr)
{
}

static unsigned int masc_tty_poll(struct tty_struct *tty, struct file *filp,
		poll_table * wait)
{
}

#endif

static int masc_tty_ioctl(struct tty_struct *tty, struct file *file,
		unsigned int cmd, unsigned long arg)
{
	masc_dev_t *masc = tty->disc_data;

	if (!masc || !masc->magic == MASC_MAGIC || !masc->flg_occup)
		return -ENXIO;
	
	if (!capable(CAP_NET_ADMIN))	
		return -EPERM;
	
	switch (cmd) {
	case MASCIOCTURNOFF:
		return masc_modem_off(masc);
	case MASCIOCTXOFF:
		return masc_transmitter_off(masc);
	case MASCIOCTXON:
		return masc_transmitter_on(masc);
	case MASCIOCSNETID:
		{
			unsigned long id;
			if (copy_from_user(&id, (void *) arg, sizeof id))
				return -EFAULT;
			return masc_set_netid(masc, id);
		}
	case MASCIOCDRAIN:
		return masc_wait_drain(masc);
	case TCGETS:
	case TCSETSF:
	case TCSETSW:
	case TCSETS:
	case TCGETA:
	case TCSETAF:
	case TCSETAW:
	case TCSETA:
		return n_tty_ioctl(tty, file, cmd, arg);
	default:
		return -ENOIOCTLCMD;
	}
}



/*
 * This function is called from the tty system to pass data to us. We
 * are given a count of bytes, and a pointer to the data (cp). Also, there
 * is a second pointer (fp) which points to an array of bytes which indicate
 * indicates which bytes have framing (parity) errors.
 * We assume that whenever we see a carriage return, we have received
 * a MASC frame, so we call the frame processing function. When this
 * happens, the receive buffer is cleared and we start again.
 * This isn't right, strictly speaking, since we should scan for
 * the leading ^ which kicks off a MASC frame.
 */

static void masc_tty_receive_buf(struct tty_struct *tty, const unsigned char *cp,
		char *fp, int count)
{
	masc_dev_t *masc = tty->disc_data;

	masc_lock(masc);

	while (count--) {
		unsigned int ch = *cp++;
		if (fp && *fp++) {
			if (!masc->flg_toss) {
				masc->flg_toss = 1;
				masc->net_stats.rx_errors++;
			}
			continue;
		}

		if (masc->rx_count < MASC_BIG_BUF_SIZE) {
			masc->rx_buf[masc->rx_count++] = ch;
		} else {
			if (!masc->flg_oflow) {
				masc->flg_oflow = 1;
				masc->net_stats.rx_errors++;
				masc_unlock(masc);
				printk(KERN_ERR MNAME
				"receive buffer overflow (%s)---truncating and tossing packet\n", masc->if_name);
				masc_lock(masc);
			}
		}

		if (ch == '\r')
			masc_process_frame(masc);
	}

	masc_unlock(masc);
}

/*
 * The tty system uses this function to query how much buffer space we have. We
 * simply say that we can take a lot. This is true, since we remove and stuff
 * it into a dynamically allocated packets that go to the network layer.
 */

static int masc_tty_receive_room(struct tty_struct *tty)
{
	return 65536;
}

/*
 * Transmit completion handler, called by the tty subsystem to indicate that it
 * has completed a write operation. Our response is to stuff more data into the
 * tty and call the tty's write method to continue the flow. If no data is
 * available, as indicated by masc_fill_transmit_buf() returning zero, we lower
 * the device busy transmitting flag, and set a flag for the bottom half
 * processing to tell the network layer to try sending us more stuff.
 */

static void masc_tty_write_wakeup(struct tty_struct *tty)
{
	masc_dev_t *masc = tty->disc_data;

	if (!masc || !masc->magic == MASC_MAGIC)
		return;

	masc_lock(masc);
	MASC_DBG_PRINTK((KERN_INFO MNAME "%s: write wakeup\n", masc->if_name));
	masc_kick_tty(masc);
	masc_unlock(masc);
}

/*
 * Generate the /proc/net/mobitex_masc data.
 */

static int masc_get_info(char *buffer, char **start, off_t offset, int length, int dummy)
{
	int len = 0;
	masc_dev_t *masc;

	len += sprintf(buffer, "Name     Batt Sign Roam Stop LastFourErr LocalMAN BcastMAN Notr Illg Cong Errs\n");

	masc_global_lock();

	for (masc = masc_list.next; masc != &masc_list; masc = masc->next) {
		struct device *dev = masc->dev;

		/*
		 * Need to disable interrupts here to get a consistent snapshot
		 * of the data. start_bh_atomic() isn't good enough since
		 * masc_tty_receive_buf is a real interrupt.
		 */

		masc_lock(masc);

		if (masc->flg_occup) {
			len += sprintf(buffer + len, "%-8s %4d %4d %4s %4s %02X,%02X,%02X,%02X "
				"%02X:%02X:%02X %02X:%02X:%02X %4lu %4lu %4lu %4lu\n",
				masc->if_name, masc->batt, masc->sign,
				masc->flg_nocontact ? "Yes" : "No",
				masc->flg_tstop ? "Yes" : "No",
				(unsigned int) masc->errhist[0], (unsigned int) masc->errhist[1],
				(unsigned int) masc->errhist[2], (unsigned int) masc->errhist[3],
				(unsigned int) dev->dev_addr[0], (unsigned int) dev->dev_addr[1],
				(unsigned int) dev->dev_addr[2],
				(unsigned int) dev->broadcast[0], (unsigned int) dev->broadcast[1],
				(unsigned int) dev->broadcast[2],
				masc->rx_notran, masc->rx_illegal, masc->rx_congest, masc->rx_errcount);
		}

		masc_unlock(masc);

		if (len > 4000)
			break;
	}

	masc_global_unlock();

	*start = buffer + offset; 
	len -= offset;

	if (len < 0)
		len = 0;

	if (len>length)
		len = length; 

	return len;
}

/*
 * Initialize the MASC subsystem, and register it as a TTY line discipline
 * module. Subsequent opening and closing of MASC devices is initiated from the
 * TTY subsystem, when a TTY is instructed to register MASC as a line
 * discipline.  When a TTY is told to install MASC line discipline, it opens a
 * MASC device, A MASC device instance is allocated and registers itself as a
 * network driver. This creates a bridge between the TTY and the kernel's
 * networking subsystem, with MASC device serving as an intermediary. Neat, huh?
 */

static int masc_do_init(void)
{
	int status = 0;	

	printk(KERN_INFO MNAME "version %s\n", MASC_VERSION);

	if (masc_maxdev < 4) {
		printk(KERN_INFO MNAME "masc_maxdev too small!\n");
		printk(KERN_INFO MNAME "was %d\n", masc_maxdev);
		masc_maxdev = 4;
		printk(KERN_INFO MNAME "set to %d\n", masc_maxdev);
	}

#ifdef CONFIG_PROC_FS
	masc_proc.namelen = sizeof "mobitex_masc" - 1;
	masc_proc.name = "mobitex_masc";
	masc_proc.mode = S_IFREG | S_IRUGO;
	masc_proc.nlink = 1;
	masc_proc.uid = 0;
	masc_proc.gid = 0;
	masc_proc.size = 0;
	masc_proc.ops = &proc_net_inode_operations;
	masc_proc.get_info = masc_get_info;
	masc_proc.fill_inode = NULL;
	masc_proc.next = NULL;
	masc_proc.parent = NULL;
	masc_proc.subdir = NULL;
	masc_proc.data = NULL;
	masc_proc_flag = proc_register(proc_net, &masc_proc);
	if (masc_proc_flag < 0) {
		printk(KERN_INFO MNAME "couldn't register /proc entry\n");
	}
#endif

	masc_list.next = &masc_list;
	masc_list.prev = &masc_list;

	masc_ldisc.magic = TTY_LDISC_MAGIC;
	masc_ldisc.name = "n_masc";
	masc_ldisc.open	= masc_tty_open;
	masc_ldisc.close = masc_tty_close;
	masc_ldisc.flush_buffer = NULL;	
	masc_ldisc.chars_in_buffer = NULL;
	masc_ldisc.read	= NULL;	
	masc_ldisc.write = NULL;
	masc_ldisc.ioctl = masc_tty_ioctl;
	masc_ldisc.poll = NULL; 
	masc_ldisc.set_termios = NULL;
	masc_ldisc.receive_room = masc_tty_receive_room;
	masc_ldisc.receive_buf = masc_tty_receive_buf;
	masc_ldisc.write_wakeup = masc_tty_write_wakeup;

	if ((status = tty_register_ldisc(N_MASC, &masc_ldisc)) == 0) {
		printk(KERN_INFO MNAME "registered line discipline\n");
	} else {
		printk(KERN_ERR MNAME "cannot register line discipline (error = %d)\n", status);
		return status;
	}

	spin_lock_init(&masc_spinlock);

	return status;
}


/*
 * If this is built as a module, the init_module() and cleanup_module()
 * functions are used as entry points to initialize and deinitialize the MASC
 * subsystem.
 *
 * On the other hand, if this is statically linked to the kernel, the single
 * function masc_init() is provided for the kernel to call at system
 * initialization time. A static driver persists until the operating system is
 * terminated, so a matching cleanup function is not needed
 */

#ifdef MODULE

int init_module(void)
{
	return masc_do_init();
}

void cleanup_module(void)
{
	int status = 0;
	masc_dev_t *masc, *next;

	if (tty_register_ldisc(N_MASC, NULL) == 0) {
		printk(KERN_INFO
			MNAME "line discipline successfully unregistered\n");
	} else {
		printk(KERN_ERR
			MNAME "unable to unregister MASC line discipline "
			"(error = %d)\n", status);
	}

#ifdef CONFIG_PROC_FS
	if (masc_proc_flag == 0)
		proc_unregister(proc_net, masc_proc.low_ino);
#endif

	for (masc = masc_list.next; masc != &masc_list; masc = next) {
		next = masc->next;
		kfree(masc);
	}

	printk(KERN_INFO MNAME "module cleaned up\n");
}

#else	/* if !defined MODULE */

int masc_init(struct device *not_used)
{
	int status = masc_do_init();
	return (status == 0) ? ENODEV : status;
}

#endif	/* !defined MODULE */
